@@ -34,6 +34,8 @@ from sklearn.metrics import confusion_matrix
 from sklearn.model_selection import ParameterGrid
 from tensorflow.keras.models import load_model
 
+from wrapt_timeout_decorator import timeout
+
 from .models import ModelsLoader
 
 class GridSearch:
@@ -211,7 +213,8 @@ class GridSearch:
         refit=False,
         clf_algorithm = '',
         clf_model_kwargs={},
-        params_generators = []
+        params_generators = [],
+        timeout_fit=None
 
     ):
         self.base_clf = clf
@@ -220,9 +223,11 @@ class GridSearch:
         self.datadir = datadir
         self.n_cpus = n_cpus
         self.clf_type = clf_type
+        self.timeout_fit = timeout_fit
 
         self.params = params        # Dictionary with params settings
         self.params_grid = []       # Will hold the actual parameter configurations returned by ParameterGrid
+        self.params_grid_noremove = []  # All the params for the current session, including the cached ones from the datadir
         self.refit = refit
 
         self.clf_name = clf.__class__.__name__
@@ -267,6 +272,9 @@ class GridSearch:
         y = params.pop("targets")
         start = time.time()
 
+        experiment_file_params = self._experiment_file_params(params)
+        logger.info(f"Fitting {experiment_file_params}")
+
         if self.clf_type == "tensorflow":  # tf has no clone
             clf = self.base_clf
         else:
@@ -305,85 +313,6 @@ class GridSearch:
             )
         return metrics
 
-    @staticmethod
-    def _test_clf_static(params):
-        """Static version of _test_clf(). Instantiates tensorflow models internally.
-
-        Args:
-            params (dict): Dictionary with all necessary parameters. Includes parameters for instantiating a tensorflow model.
-
-        Returns:
-            dict: Metrics dict
-        """
-        X = params.pop("samples")
-        y = params.pop("targets")
-        clf_algorithm = params.pop("clf_algorithm")
-        clf_model_kwargs = params.pop("clf_model_kwargs")
-        clf_type = params.pop("clf_type")
-        datadir = params.pop("datadir")
-        clf_name = params.pop("clf_name")
-        experiment_file_params = params.pop("experiment_file_params")
-        start = time.time()
-
-        logger.info(f"Running {experiment_file_params}")
-
-        # Assume self.clf_type == "tensorflow"
-        clf = ModelsLoader.get(clf_algorithm, **clf_model_kwargs)
-
-        clf = clf.set_params(**params)
-        y_pred = clf.fit_predict(X)
-
-        metrics = _fpfn(None, None, y_pred, y)
-        end = time.time()
-        duration = end - start
-        metrics["duration"] = duration
-
-        logger.debug(metrics)
-
-        fname = f"{datadir}/{clf_name}" + experiment_file_params
-
-        if clf_type == "tensorflow":  # tf is not pickle compatible
-
-            # Save model
-            clf.metrics_ = metrics
-            clf.save(fname)
-
-            # Manually pickle attributes in a separate file
-            attrs_to_pickle = clf.get_config()
-            attrs_to_pickle['optimizer'] = params['optimizer']
-            attrs_to_pickle['loss'] = params['loss']
-            pickle.dump(attrs_to_pickle, open(os.path.join(fname, 'autoencoder_attrs'),'wb'))
-
-            # Pickle metrics
-            pickle.dump(
-                # [ba, tpr, tnr, duration, tn, fp, fn, tp],
-                metrics,
-                open(
-                    fname + "/metrics",
-                    "wb",
-                ),
-            )
-
-            # Save metrics also as a text file
-            with open(os.path.join(fname, 'metrics.txt'), "w") as external_file:
-                pprint.pprint(metrics, stream=external_file)
-
-            # Save config to a text file
-            with open(os.path.join(fname, 'config.txt'), "w") as external_file:
-                pprint.pprint(clf.get_config(), stream=external_file)
-
-        else:
-            metrics["clf"] = clf
-            pickle.dump(
-                # [clf, ba, tpr, tnr, duration, tn, fp, fn, tp],
-                metrics,
-                open(
-                    fname,
-                    "wb",
-                ),
-            )
-        return metrics
-
     def _grid_make_iterable(self):
         for k, v in self.params.items():
             if not isinstance(self.params[k], Iterable):
@@ -412,6 +341,10 @@ class GridSearch:
         # Generators can overwrite values defined in self.params
         self.params_grid = [_merge_dict_tuple(tup) for tup in itertools.product(self.params_grid, *self.params_generators)]
 
+        # Make a copy of the full list
+        # Cached estimators found in the save folder will be removed from params_grid, but are kept in params_grid_all
+        self.params_grid_noremove = self.params_grid
+
     def _experiment_get_path(self, prefix, p):
         return prefix + self._experiment_file_params(p)
 
@@ -422,12 +355,15 @@ class GridSearch:
     def _experiment_remove_if_exists(self, prefix, p):
         experiment = self._experiment_get_path(prefix, p)
         if os.path.exists(experiment):
-            shutil.rmtree(experiment)            
+            if os.path.isfile(experiment):
+                os.remove(experiment)
+            else:
+                shutil.rmtree(experiment)
 
     def _experiment_file_params(self, p):
         str_params = ""
         for k, v in p.items():
-            if k == "samples" or k == "targets" or k == "savename_key_len":
+            if k == "samples" or k == "targets" or k == "savename_key_len" or k == "timeout_fit":
                 continue
             if type(v) == np.ndarray:
                 v = self._params_array_subsample_hash(v)
@@ -469,7 +405,53 @@ class GridSearch:
             prefix + self._experiment_file_params(p) for p in self.params_in_
         ]
         for i, experiment in enumerate(experiments):
+            if os.path.exists(experiment):
+                logging.info(f"{experiment}")
+                if os.path.isdir(experiment):  # tf
+                    clf = load_model(experiment)
+
+                    # Restore model, including the manually saved attributes
+                    with open(os.path.join(str(experiment), "autoencoder_attrs"), "rb") as fp:
+                        clf_attrs = pickle.load(fp)
+                    # Change object's class to actual Autoencoder, so methods are available. Weird but works.
+                    clf.__class__ = ModelsLoader.get_constructor(self.clf_algorithm)
+                    # Update with the manually pickled attributes
+                    clf.set_params(**clf_attrs)
+
+                    with open(str(experiment) + "/metrics", "rb") as fp:
+                        metrics = pickle.load(fp)
+                else:
+                    with open(experiment, "rb") as fp:
+                        metrics = pickle.load(fp)
+                    clf = metrics["clf"]
+                self._compare_results(experiment, clf, metrics)
+
+        params = str(self.best_experiment_).split("-")[1:]
+        params = ",".join(params)
+        durationstr = f"{self.best_duration_:.2f}" if self.best_duration_ else f"{self.best_duration_}"
+        logging.info(
+            f"{self.clf_name.upper()} BEST [" + durationstr +"s]"  
+            f"({params}): "
+            f"{self.best_score_}"
+        )
+
+        if hasattr(self.best_estimator_, "get_params"):  # sklearn
+            self.best_params_ = self.best_estimator_.get_params()
+            self.labels_ = self.best_estimator_.labels_
+
+    def _grid_get_estimators(self):
+        p = Path(self.datadir)
+        
+        # Read saved estimators from folder
+        experiments = list(p.glob(f"{self.clf_name}-*"))
+
+        # Keep only the experiments from the current session, match by path name
+        exp_names_current_session = [ os.path.join(p, f"{self.clf_name}") + self._experiment_file_params(e) for e in self.params_grid_noremove]
+        experiments = [e for e in experiments if str(e) in exp_names_current_session]
+
+        for experiment in experiments:
             logging.info(f"{experiment}")
+
             if os.path.isdir(experiment):  # tf
                 clf = load_model(experiment)
 
@@ -483,35 +465,17 @@ class GridSearch:
 
                 with open(str(experiment) + "/metrics", "rb") as fp:
                     metrics = pickle.load(fp)
+
+                self.estimators_.append(clf)
+
             else:
                 with open(experiment, "rb") as fp:
+                    # [clf, ba, tpr, tnr, duration, tn, fp, fn, tp] = pickle.load(fp)
                     metrics = pickle.load(fp)
-                clf = metrics["clf"]
-            self._compare_results(experiment, clf, metrics)
-
-        params = str(self.best_experiment_).split("-")[1:]
-        params = ",".join(params)
-        logging.info(
-            f"{self.clf_name.upper()} BEST [{self.best_duration_:.2f}s]"
-            f"({params}): "
-            f"{self.best_score_}"
-        )
-
-        if hasattr(self.best_estimator_, "get_params"):  # sklearn
-            self.best_params_ = self.best_estimator_.get_params()
-            self.labels_ = self.best_estimator_.labels_
-
-    def _grid_get_estimators(self):
-        p = Path(self.datadir)
-        experiments = list(p.glob(f"{self.clf_name}-*"))
-        for experiment in experiments:
-            with open(experiment, "rb") as fp:
-                # [clf, ba, tpr, tnr, duration, tn, fp, fn, tp] = pickle.load(fp)
-                metrics = pickle.load(fp)
-                clf = metrics["clf"]
-                self.estimators_.append(clf)
-                print(experiment)
-                print(clf)
+                    clf = metrics["clf"]
+                    self.estimators_.append(clf)
+                    print(experiment)
+                    print(clf)
 
     def _grid_find_mean_std(self):
         bas = []
@@ -532,41 +496,50 @@ class GridSearch:
         )
 
     def _grid_search(self, X, y):
-        self._grid_generate_search_space(X, y)
-        if self.refit:
-            self._grid_remove_existing_experiments()
-        else:
-            self._grid_remove_existing_tests()
 
-        if self.n_cpus == 1:
-            for p in self.params_grid:
-                self._test_clf(p)
+        # Try to retrieve a timeout for trainning each estimator
+        # Individual algorithm value (self.params), if specified, has precedence over global value self.timeout_fit)
+        timeout_fit = self.params['timeout_fit'][0] if 'timeout_fit' in self.params else self.timeout_fit
 
+        # Append new parameters so that _test_clf_static() can instantiate the model internally
+        new_params = {  "clf_algorithm": self.clf_algorithm,
+                        "clf_model_kwargs": self.clf_model_kwargs,
+                        "clf_type": self.clf_type,
+                        "datadir": self.datadir,
+                        "clf_name": self.clf_name,
+                        "timeout_fit": timeout_fit
+        }
+        # Merge new params with each set in the existing list
+        params_grid_augmented = []
+        for p in self.params_grid:
+            p_clone = p.copy()  # Make a clone of the parameters dict. In case of timeout error, added params are not popped, and pollute the original params list
+            p_clone.update({"experiment_file_params": self._experiment_file_params(p)})
+            p_clone.update(new_params)
+            params_grid_augmented.append(p_clone)
+
+        # For tf-based models, call _test_clf_static() instead of _test_clf
         # tf incompatible with mp according to tf issue #46316
-        elif self.clf_type == "tensorflow":
+        if self.clf_type == "tensorflow":
+            for p in params_grid_augmented:
 
-            # Append new parameters so that _test_clf_static() can instantiate the model internally
-            new_params = {  "clf_algorithm": self.clf_algorithm,
-                            "clf_model_kwargs": self.clf_model_kwargs,
-                            "clf_type": self.clf_type,
-                            "datadir": self.datadir,
-                            "clf_name": self.clf_name,
-            }
-            for p in self.params_grid:
-                p.update({"experiment_file_params": self._experiment_file_params(p)})
-                p.update(new_params)
+                # tf-based models cannot be run with timeout, because of incompatibility with mp
+                del p['timeout_fit']
 
-                self._test_clf_static(p)    
+                _test_clf_static(p)    
 
             # Still can't parallelize, models get instantiated but hang when fitting()
             # pool_obj = multiprocessing.Pool(self.n_cpus)
-            # pool_obj.map(self._test_clf_static, self.params_grid)
+            # pool_obj.map(_test_clf_static, params_grid_augmented)
+
+        elif self.n_cpus == 1:
+            for p in params_grid_augmented:
+                _test_clf_static(p)    
 
         else:
             pool_obj = multiprocessing.Pool(self.n_cpus)
-            pool_obj.map(self._test_clf, self.params_grid)
+            pool_obj.map(_test_clf_static, params_grid_augmented)
 
-    def fit(self, X, y=None):
+    def fit(self, X, y=None, only_cached=False):
         """Perform grid search on samples in `X` with the selected `clf`.
 
         Parameters
@@ -584,17 +557,36 @@ class GridSearch:
             Fitted estimator.
         """
 
-        self._grid_search(X, y)
+        # Generate parameters
+        self._grid_generate_search_space(X, y)
+        if self.refit:
+            self._grid_remove_existing_experiments()
+        else:
+            self._grid_remove_existing_tests()
+
+        # Actual search
+        if not only_cached:
+            self._grid_search(X, y)
+
         if y is None:
             self._grid_get_estimators()
         else:
             self._grid_find_best_result()
+        #self.load(find_best=(y is not None))
 
     # def _test_rounds(self, X, y):
     #     self._grid_search(X, y)
     #     self._grid_find_mean_std()
 
 
+    # def load(self, find_best=False):
+    #     """Only loads existing (cached) estimators from disk
+    #     """
+    #     if find_best:
+    #         self._grid_find_best_result()
+    #     else:
+    #         self._grid_get_estimators()
+
 def _fpfn(cls, clf, y_pred, y_true):
     if y_true is None:
         metrics = {
@@ -632,3 +624,101 @@ def _compare_metric_balanced_accuracy(cls, experiment, clf, metrics):
     if metrics["ba"] > cls.best_score_["ba"]:
         return True
     return False
+
+def _test_clf_static(params):
+    """Function version of _test_clf(), outside of class definition. Internally instantiates models from scratch.
+
+    Args:
+        params (dict): Dictionary with all necessary parameters, including the ones for instantiating models.
+
+    Returns:
+        dict: Metrics dict
+    """
+    X = params.pop("samples")
+    y = params.pop("targets")
+    clf_algorithm = params.pop("clf_algorithm")
+    clf_model_kwargs = params.pop("clf_model_kwargs")
+    clf_type = params.pop("clf_type")
+    datadir = params.pop("datadir")
+    clf_name = params.pop("clf_name")
+    experiment_file_params = params.pop("experiment_file_params")
+    if 'timeout_fit' in params:
+        timeout_fit = params.pop('timeout_fit')
+    else:
+        timeout_fit = None
+    
+    start = time.time()
+    logger.info(f"Fitting {clf_algorithm} with timeout={timeout_fit}: {experiment_file_params}")
+
+    # Create model from scratch and parameterize
+    clf = ModelsLoader.get(clf_algorithm, **clf_model_kwargs)
+    clf = clf.set_params(**params)
+
+    if timeout_fit is not None:
+
+        # Local function to replace clf.fit_predict()
+        # because we can't apply timeout() decorator directly to a function with side-effects
+        def clf_fit_predict(clf, X):
+            y_pred = clf.fit_predict(X)
+            return y_pred, clf
+
+        # Fit and catch a possible TimeoutError
+        try:
+            y_pred, clf = timeout(dec_timeout=timeout_fit, use_signals=False)(clf_fit_predict)(clf, X)
+        except TimeoutError as e:
+            logger.info(f"Timeout exceeded! : {experiment_file_params}")
+            return None
+    else:
+        y_pred = clf.fit_predict(X)
+    logger.info(f"Finished {clf_algorithm} {experiment_file_params}")
+
+    metrics = _fpfn(None, None, y_pred, y)
+    end = time.time()
+    duration = end - start
+    metrics["duration"] = duration
+
+    logger.debug(metrics)
+
+    fname = f"{datadir}/{clf_name}" + experiment_file_params
+
+    if clf_type == "tensorflow":  # tf is not pickle compatible
+
+        # Save model
+        clf.metrics_ = metrics
+        clf.save(fname)
+
+        # Manually pickle attributes in a separate file
+        attrs_to_pickle = clf.get_config()
+        attrs_to_pickle['optimizer'] = params['optimizer']
+        attrs_to_pickle['loss'] = params['loss']
+        pickle.dump(attrs_to_pickle, open(os.path.join(fname, 'autoencoder_attrs'),'wb'))
+
+        # Pickle metrics
+        pickle.dump(
+            # [ba, tpr, tnr, duration, tn, fp, fn, tp],
+            metrics,
+            open(
+                fname + "/metrics",
+                "wb",
+            ),
+        )
+
+        # Save metrics also as a text file
+        with open(os.path.join(fname, 'metrics.txt'), "w") as external_file:
+            pprint.pprint(metrics, stream=external_file)
+
+        # Save config to a text file
+        with open(os.path.join(fname, 'config.txt'), "w") as external_file:
+            pprint.pprint(clf.get_config(), stream=external_file)
+
+    else:
+        metrics["clf"] = clf
+        pickle.dump(
+            # [clf, ba, tpr, tnr, duration, tn, fp, fn, tp],
+            metrics,
+            open(
+                fname,
+                "wb",
+            ),
+        )
+    return metrics
\ No newline at end of file
