@@ -14,6 +14,9 @@
 # ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 # OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 
+import logging
+logger = logging.getLogger(__name__)
+
 import numpy as np
 import tensorflow as tf
 from sklearn.preprocessing import MinMaxScaler
@@ -24,19 +27,21 @@ from tensorflow.keras.callbacks import EarlyStopping, ReduceLROnPlateau
 
 @tf.keras.utils.register_keras_serializable()
 class Autoencoder(tf.keras.Model):
-    """Autoencoder model introduced by Aggarwal C. [1]_
+    """Autoencoder model introduced by Aggarwal C. [Aggarwal2015]_
 
     Parameters
     ----------
     encoder_neurons : list
         The number of neurons per encoder layers.
+        Values > 1 are absolute, values <= 1 are relative to input size (e.g. [0.8, 0.4])
 
     decoder_neurons : list
         The number of neurons per decoder layers.
+        Values > 1 are absolute, values <= 1 are relative to input size (e.g. [0.8, 0.4])
 
     activation_function : str or callable, default='tanh'
         The activation function for all layers.
-        See `Keras doc <https://keras.io/api/layers/activations/>`_.
+        See `Keras doc <https://keras.io/api/layers/activations/>`__.
 
     l2_regularizer : float in (0., 1), default=0.1
         The regularization factor for L2 regularizer for all layers.
@@ -52,11 +57,73 @@ class Autoencoder(tf.keras.Model):
         The contamination rate used for computing the reconstruction error
         threshold. Used if threshold is not set.
 
+    optimizer: str or callable, default='adam'
+        The optimizer algorithm of the neural network.
+        See `Keras doc <https://keras.io/api/optimizers/>`__.
+
+    loss: str or callable, default='mse'
+        The loss function of the neural network.
+        See `Keras doc <https://keras.io/api/losses/>`__.
+
+    epochs: int, default=1000
+        Maximum number of epochs for network training.
+        See arguments of fit() method in `Keras doc <https://keras.io/api/models/model_training_apis/>`__.
+
+    batch_size: int, default=10000
+        Batch size used for training.
+        See arguments of fit() method in `Keras doc <https://keras.io/api/models/model_training_apis/>`__.
+
+    shuffle: boolean, default=True
+        Shuffle data before each epoch during training.
+        See arguments of fit() method in `Keras doc <https://keras.io/api/models/model_training_apis/>`__.
+
+    validation_size: float, default==0.1
+        Amount of training set used for validation.
+        See `validation_split` argument of fit() method in `Keras doc <https://keras.io/api/models/model_training_apis/>`__.
+
+    learning_rate : float, default==0.001
+        Gradient step size used in training.
+        Depends on the optimizer used.
+
+    verbose: int or str, default='auto'
+        Print training progress. Can be 0, 1, or 'auto'
+        See arguments of fit() method in `Keras doc <https://keras.io/api/models/model_training_apis/>`__.
+
+    es_monitor: str or None, default='val_loss'
+        Quantity to monitor for early stopping of training.
+        Use None to disable EarlyStopping.
+        See `Keras doc <https://keras.io/api/callbacks/early_stopping/>`__.
+
+    es_patience: int, default=50
+        Early Stopping: how many epochs to wait before stop.
+        See `Keras doc <https://keras.io/api/callbacks/early_stopping/>`__.
+
+    redlr_monitor: str or None, default='val_loss'
+        Quantity to monitor for "Reduce Learning Rate on Plateau" during training.
+        Use None to disable ReduceLROnPlateau.
+        See `Keras doc <https://keras.io/api/callbacks/reduce_lr_on_plateau/>`__.
+
+    redlr_learning_rate_reduce_factor: float, default=0.2
+        Factor for reducing learning rate when plateau is detected. Must be < 1.
+        Use None to disable ReduceLROnPlateau.
+        See `Keras doc <https://keras.io/api/callbacks/reduce_lr_on_plateau/>`__.
+
+    redlr_learning_rate_min: float, default=0.00001
+        Minimum learning rate allowed when reducing learning rate on plateau.
+        See `Keras doc <https://keras.io/api/callbacks/reduce_lr_on_plateau/>`__.
+
+    redlr_patience: int, default=10
+        How many epochs to wait before reducing learning rate on plateau.
+        See `Keras doc <https://keras.io/api/callbacks/reduce_lr_on_plateau/>`__.
+
+    savename_key_len: int, default=3
+        When saving the model, restrict key length in the name to this many characters.
+
     Attributes
     ----------
-    reconstruction_errors_ : array-like of shape (n_samples,)
-        The raw outlier scores for the training data. The anomalies have
-        a larger error score.
+    decision_scores_ : array-like of shape (n_samples,)
+        The raw outlier scores for the training data, i.e. the reconstruction errors.
+        The anomalies have a larger error score.
 
     history_ : Keras Object
         The training history of the model.
@@ -68,77 +135,129 @@ class Autoencoder(tf.keras.Model):
     threshold_ : float
         The threshold for the raw outliers scores.
 
+    encoder_neurons_ : list
+        The actual number of neurons per encoder layers, if decoder_neurons is relative
+
+    decoder_neurons_ : list
+        The actual number of neurons per decoder layers, if decoder_neurons is relative
+
     References
     ----------
-    .. [1] Aggarwal C. (2015) Outlier Analysis. In: Data Mining. Springer,
+    .. [Aggarwal2015] Aggarwal C. (2015) Outlier Analysis. In: Data Mining. Springer,
         Cham. https://doi.org/10.1007/978-3-319-14142-8_8
     """
 
     def __init__(
         self,
-        encoder_neurons=None,
-        decoder_neurons=None,
-        activation_function="tanh",
-        l2_regularizer=0.1,
-        dropout=0.2,
-        threshold=None,
-        contamination=0.1,
-        optimizer=None,
-        loss=None,
-        epochs=None,
-        batch_size=None,
-        shuffle=None,
-        validation_size=None,
+        encoder_neurons = None,
+        decoder_neurons = None,
+        activation_function = "relu",
+        l2_regularizer = 0.1,
+        dropout = 0.5,
+        threshold = None,
+        contamination = 0.1,
+        optimizer = 'adam',
+        loss = "mse",
+        epochs = 1000,
+        batch_size = 10000,
+        shuffle = True,
+        validation_size = 0.1,
+        learning_rate = 0.001,
+        verbose = 1,
+        es_monitor = 'val_loss',                       # What to monitor. Required. Comment-out to disable EarlyStopping
+        es_patience = 50,                              # How many epochs to wait before stop. Defaults to 0.
+        redlr_monitor = 'val_loss',                    # What to monitor. Required. Comment-out to disable ReduceLROnPlateau
+        redlr_learning_rate_reduce_factor = 0.2,       # Reduce factor. Must be < 1. Required. Comment-out to disable ReduceLROnPlateau
+        redlr_learning_rate_min = 0.00001,             # Minimum learning rate. Defaults to 0.
+        redlr_patience = 10,                           # How many epochs to wait before reducing. Defaults to 10.
+        savename_key_len = 3                           # When saving, restrict key length in the name to this many characters
     ):
         super(Autoencoder, self).__init__()
 
+        # Fixed parameters, passed via __init__()
         self.encoder_neurons = encoder_neurons
         self.decoder_neurons = decoder_neurons
         self.activation_function = activation_function
         self.l2_regularizer = l2_regularizer
         self.dropout = dropout
+        self.threshold = threshold              # Fixed threshold parameter specified at instantiation, while threshold_ is the actual threshold to use
         self.contamination = contamination
-        self.threshold = threshold
-        self.classes = 2
-        self.reconstruction_errors_ = None
-        self.history_ = None
-        self.labels_ = None
-
         self.optimizer = optimizer
         self.loss = loss
         self.epochs = epochs
         self.batch_size = batch_size
         self.shuffle = shuffle
         self.validation_size = validation_size
+        self.learning_rate = learning_rate
+        self.verbose = verbose
+        self.es_monitor = es_monitor
+        self.es_patience = es_patience
+        self.redlr_monitor = redlr_monitor
+        self.redlr_learning_rate_reduce_factor = redlr_learning_rate_reduce_factor
+        self.redlr_learning_rate_min = redlr_learning_rate_min
+        self.redlr_patience = redlr_patience
+        self.savename_key_len = savename_key_len
 
-        self._check_parameters()
-        self._build_model()
+        # Fixed parameters, internal
+        self.classes = 2
+
+        # Attributes estimated from data, ending with `_`
+        # See here for a general discussion: https://scikit-learn.org/stable/developers/develop.html
+        self.threshold_ = None                  # Threshold set based on contamination level, if self.threshold is None, else equal to threshold.
+        self.decision_scores_ = None
+        self.history_ = None
+        self.labels_ = None
+        self.encoder_neurons_ = None    # Actual layer sizes, if case encoder_neurons contains relative values
+        self.decoder_neurons_ = None    # Actual layer sizes, if case decoder_neurons contains relative values
+
+        # sk-learn: nothing should be done in constructor, not even validation
+        # validation and building the Keras model is done in fit(), after set_params()
 
     # See here: https://www.tensorflow.org/api_docs/python/tf/keras/utils/register_keras_serializable
     # "To be serialized and deserialized, classes must implement the get_config() method."
     def get_config(self):
-        return {"encoder_neurons": self.encoder_neurons,
-                "decoder_neurons": self.decoder_neurons,
-                "activation_function": self.activation_function,
-                "l2_regularizer": self.l2_regularizer,
-                "dropout": self.dropout,
-                "threshold": self.threshold,
-                "contamination": self.contamination,
-                "classes": self.classes,
-                "reconstruction_errors_": self.reconstruction_errors_,
-                "history_": self.history_,
-                "labels_": self.labels_,
-                "optimizer": self.optimizer,
-                "loss": self.loss,
-                "epochs": self.epochs,
-                "batch_size": self.batch_size,
-                "shuffle": self.shuffle,
-                "validation_size": self.validation_size
-        }
+        """Returns all data members of the object, i.e. both the parameters and the attributes (ending in `_`)
+
+        Needed for serialization / deserialization.
+        """
+        # Get object attributes
+        attributes = {  "classes": self.classes,
+                        "threshold_": self.threshold_,
+                        "decision_scores_": self.decision_scores_,
+                        "history_": self.history_,
+                        "labels_": self.labels_,
+                        "encoder_neurons_": self.encoder_neurons_,
+                        "decoder_neurons_": self.decoder_neurons_
+            }
+
+        # Join the parameters with the attributes in a single dict
+        params = self.get_params()
+        params.update(attributes)
+        return params
+
+    @classmethod
+    def from_config(cls, config):
+        """Instantiates an object based on a config dict returned by get_config()
+        """
+
+        # Extract attributes from `config` and set later, they cannot be passed to __init__()
+        attrs = ["classes", "threshold_", "decision_scores_", "history_", "labels_", "encoder_neurons_", "decoder_neurons_"]
+        attrs_dict = {attr: config.pop(attr) for attr in attrs}
+
+        # Instantiate with the remaining keys
+        instance = cls(**config)
+
+        # Set attributes
+        for attr, value in attrs_dict.items():
+            setattr(instance, attr, value)
+
+        return instance
 
     def _build_model(self):
+
         encoder_layers = []
-        for neurons in self.encoder_neurons:
+
+        for neurons in self.encoder_neurons_:
             linear_layer = layers.Dense(
                 neurons,
                 activation=self.activation_function,
@@ -150,11 +269,11 @@ class Autoencoder(tf.keras.Model):
             encoder_layers.append(dropout_layer)
 
         decoder_layers = []
-        for ilayer, neurons in enumerate(self.decoder_neurons):
+        for ilayer, neurons in enumerate(self.decoder_neurons_):
 
             # The last layer should have no activation function
             activation = self.activation_function
-            if ilayer == len(self.decoder_neurons)-1:
+            if ilayer == len(self.decoder_neurons_)-1:
                 activation = 'linear'
 
             linear_layer = layers.Dense(
@@ -171,12 +290,14 @@ class Autoencoder(tf.keras.Model):
         self.decoder = tf.keras.Sequential(decoder_layers[:-1], name="decoder")
 
     def call(self, inputs):
-        if inputs.shape[1] != self.decoder_neurons[-1]:
+        """Runs the model on the input data.
+        """
+        if inputs.shape[1] != self.decoder_neurons_[-1]:
             raise ValueError(
                 "Expected the number of features to be equal to "
                 "the number of neurons on the last decoder layer. "
                 f"But found: {inputs.shape[1]} features and "
-                f"{self.decoder_neurons[-1]} neurons."
+                f"{self.decoder_neurons_[-1]} neurons."
             )
 
         encoded = self.encoder(inputs)
@@ -189,8 +310,8 @@ class Autoencoder(tf.keras.Model):
         print(self.encoder.summary())
         print(self.decoder.summary())
 
-    def predict_reconstruction_error(self, X):
-        """Predict the reconstruction errors for some samples.
+    def decision_function(self, X):
+        """Compute the reconstruction errors for some samples.
         The anomalies have a larger error score.
 
         This does not modify the object. Use for prediction, not fitting.
@@ -210,15 +331,12 @@ class Autoencoder(tf.keras.Model):
 
     def predict_proba(self, X=None, method="linear", thresh=None):
         """Compute a distribution probability on the reconstuction errors
-        predicted for the samples passed as parameter. The scores as
-        normalized using a scaler fitted on the train scores, so the method
-        predict_decision_scores() must be called after training the model
-        in order to be able to compute the probabilities.
+        predicted for the samples passed as parameter.
 
         This does not modify the object. Use for prediction, not fitting.
 
         If X is not None, the errors are computed for the data in X
-        If X is None, assume the erorrs are already available in self.reconstruction_errors_
+        If X is None, assume the erorrs are already available in `self.decision_scores_`
 
         Parameters
         ----------
@@ -226,11 +344,12 @@ class Autoencoder(tf.keras.Model):
             The samples for which to compute the probabilities.
 
         method : {'linear', 'hard', 'quantile'}, default='linear'
-            The method used for score normalization. 
-            'linear': scale scores to [0, 1] range
-            'hard': hard thresholding with a given threshold `thresh`, return binary scores. If `thresh` is None, use self.threshold instead.
-            'quantile': hard thresholding using the given quantile in `thresh`, 
-                        e.g. thresh=0.1 assigns 0 to the smallest 10% errors, and 1 to the largest 90% errors.
+                 The method used for score normalization. Possible values:
+
+                 - 'linear': scale scores to [0, 1] range. Default.
+                 - 'hard': hard thresholding with a given threshold `thresh`, return binary scores. If `thresh` is None, use `self.threshold_` instead.
+                 - 'quantile': hard thresholding using the given quantile in `thresh`,
+                   e.g. thresh=0.1 assigns 0 to the smallest 10% errors, and 1 to the largest 90% errors.
 
         Returns
         -------
@@ -239,13 +358,13 @@ class Autoencoder(tf.keras.Model):
             anomaly prob on dimension 1).
         """
 
-        # If X is None, we process the error values stored in self.reconstruction_errors_
-        if X is None and self.reconstruction_errors_ is None:
+        # If X is None, we process the error values stored in self.decision_scores_
+        if X is None and self.decision_scores_ is None:
             raise Exception(
                 "Reconstruction errors on train set are not available."
             )
 
-        rec_errors = self.predict_reconstruction_error(X) if X is not None else self.reconstruction_errors_
+        rec_errors = self.decision_function(X) if X is not None else self.decision_scores_
 
         probs = np.zeros([rec_errors.shape[0], self.classes])
 
@@ -254,8 +373,8 @@ class Autoencoder(tf.keras.Model):
             probs[:, 1] = scaler.fit_transform(rec_errors.reshape(-1, 1)).squeeze()
             probs[:, 0] = 1 - probs[:, 1]
 
-        elif method == "hard" and (thresh is not None or self.threshold is not None):
-            thresh_local = thresh if thresh is not None else self.threshold   # given threshold, or fitted one
+        elif method == "hard" and (thresh is not None or self.threshold_ is not None):
+            thresh_local = thresh if thresh is not None else self.threshold_   # given threshold, or fitted one
             probs[:, 1] = (rec_errors > thresh_local).astype("int")
 
         elif method == "quantile" and thresh is not None:
@@ -270,11 +389,16 @@ class Autoencoder(tf.keras.Model):
         return probs
 
     def set_params(self, **kwargs):
+        """Set estimator parameters"""
         for key, value in kwargs.items():
             setattr(self, key, value)
         return self
 
     def get_params(self, deep=True):
+        """Returns all parameters of the estimator. Needed for Scikit-learn compatibility
+        """
+
+        # Don't return the attributes (ending in `_``)
         return {
             "encoder_neurons": self.encoder_neurons,
             "decoder_neurons": self.decoder_neurons,
@@ -289,14 +413,37 @@ class Autoencoder(tf.keras.Model):
             "batch_size": self.batch_size,
             "shuffle": self.shuffle,
             "validation_size": self.validation_size,
-            #reconstruction_errors_?
-            #history_?
-            #labels_?
+            "learning_rate": self.learning_rate,
+            "verbose": self.verbose,
+            "es_monitor": self.es_monitor,
+            "es_patience": self.es_patience,
+            "redlr_monitor": self.redlr_monitor,
+            "redlr_learning_rate_reduce_factor": self.redlr_learning_rate_reduce_factor,
+            "redlr_learning_rate_min": self.redlr_learning_rate_min,
+            "redlr_patience": self.redlr_patience,
+            "savename_key_len": self.savename_key_len
         }
 
     def fit(self, X, y=None, **kwargs):
+        """Fits the model on the data provided."""
+
         self.set_params(**kwargs)
 
+        # When the last decoder layer has size 1, the values are relative.
+        # Scale the numbers by input size.
+        # Note the trailing underscore_!
+        input_size = X.shape[1]
+        if round(self.decoder_neurons[-1]) == 1:
+            self.encoder_neurons_ = [round(v * input_size) for v in self.encoder_neurons]
+            self.decoder_neurons_ = [round(v * input_size) for v in self.decoder_neurons]
+            logger.info(f'Layer size was relative, actual sizes are: encoder: {self.encoder_neurons_}, decoder: {self.decoder_neurons_}')
+        else:
+            self.encoder_neurons_ = self.encoder_neurons
+            self.decoder_neurons_ = self.decoder_neurons
+
+        self._check_parameters()
+        self._build_model()
+
         self.compile(optimizer=self.optimizer, loss=self.loss)
 
         # Overwrite learning rate, if specified
@@ -316,7 +463,7 @@ class Autoencoder(tf.keras.Model):
                 self.redlr_learning_rate_min if hasattr(self, 'redlr_learning_rate_min') else 0    # defaults to 0
             redlr_patience = \
                 self.redlr_patience if hasattr(self, 'redlr_patience') else 10   # defaults to 10
-            callbacks.append(ReduceLROnPlateau(monitor=self.redlr_monitor, factor=self.redlr_learning_rate_reduce_factor, 
+            callbacks.append(ReduceLROnPlateau(monitor=self.redlr_monitor, factor=self.redlr_learning_rate_reduce_factor,
                                             patience=redlr_patience, min_lr=redlr_learning_rate_min))
 
         # Train!
@@ -336,12 +483,14 @@ class Autoencoder(tf.keras.Model):
         )
 
         # Fit the threshold, based on the reconstruction errors on the training set
-        self.reconstruction_errors_ = self.predict_reconstruction_error(X)
+        self.decision_scores_ = self.decision_function(X)
         if self.threshold is None:
-            self.threshold = np.quantile(self.reconstruction_errors_, 1 - self.contamination)
+            self.threshold_ = np.quantile(self.decision_scores_, 1 - self.contamination)
+        else:
+            self.threshold_ = self.threshold
 
         # Compute the labels on the training set
-        self.labels_ = self.predict_proba(method='hard', thresh=self.threshold)[:,1]
+        self.labels_ = self.predict_proba(method='hard', thresh=self.threshold_)[:,1]
 
         return self
 
@@ -361,12 +510,13 @@ class Autoencoder(tf.keras.Model):
             The predicted labels for all input.
         """
 
-        pred_score = self.predict_reconstruction_error(X)
-        labels = (pred_score > self.threshold).astype("int")
+        pred_score = self.decision_function(X)
+        labels = (pred_score > self.threshold_).astype("int")
         return labels
 
     def fit_predict(self, X, y=None):
-        #return self.fit(X, y).predict(X)
+        """Run fit() and the predict()"""
+        # Fit already does prediction internally, and the labels are in labels_
         return self.fit(X, y).labels_
 
     def _check_parameters(self):
@@ -382,11 +532,18 @@ class Autoencoder(tf.keras.Model):
                 f"received {type(self.encoder_neurons)}"
             )
 
-        if not all(map(lambda x: isinstance(x, int), self.encoder_neurons)):
-            raise TypeError("Not all elements from encoder_neurons have int " "type")
+        if not all(map(lambda x: isinstance(x, int), self.encoder_neurons_)):
+            raise TypeError("Not all elements from encoder_neurons_ have int " "type")
 
-        if not all(map(lambda x: isinstance(x, int), self.decoder_neurons)):
-            raise TypeError("Not all elements from decoder_neurons have int " "type")
+        if not all(map(lambda x: isinstance(x, int), self.decoder_neurons_)):
+            raise TypeError("Not all elements from decoder_neurons_ have int " "type")
+
+        # Check layers size is not 0
+        if any (map(lambda x: x == 0, self.encoder_neurons_)):
+            raise ValueError("Some layer size from encoder_neurons_ is 0")
+
+        if any (map(lambda x: x == 0, self.decoder_neurons_)):
+            raise ValueError("Some layer size from decoder_neurons_ is 0")
 
         # Check regularizer type and value
         if not isinstance(self.l2_regularizer, float):
