@@ -276,6 +276,19 @@ def graph_to_egonet_features(
             if Gs.degree(nn) > 1:  # keep only nodes that have more than one neighbor
                 enodes.append(nn)
                 i_ego += 1
+
+        # There is a case when egonet looks like this :
+        #           degree: DiDegreeView({273172446621881.0: 3, 198112242249590.0: 1})
+        #           edges: OutEdgeView([(273172446621881.0, 273172446621881.0), (198112242249590.0, 273172446621881.0)])
+        #           in_edges: InEdgeView([(198112242249590.0, 273172446621881.0), (273172446621881.0, 273172446621881.0)])
+        #           out_edges: OutEdgeView([(273172446621881.0, 273172446621881.0), (198112242249590.0, 273172446621881.0)])
+        # And the current node is 198112242249590.0
+        # As the degree of the current node is 1, it won't be added to the enodes list
+        # This means that the reduced egonet doesn t contain the "ego" anymore
+        # So the reduced egonet is empty
+        if node not in enodes:
+            i_ego = 0
+
         # if something left in the egonet (i.e., not a star)
         if i_ego > 0:
             Gs = Gs.subgraph(enodes[0:i_ego])  # the reduced egonet
